#### qupdate.awk, copyright 2018 Charles Daniels ##############################
#
# Perform the Q-table update algorithm
#
# First read in from standard in the parameters in the following order:
#
#	x_k, x_k+1, u_k, r_k, alpha, gamma
#
# Reads an arbitrary number of records in from standard input in the format
#
#	state, action, Q
#
# Outputs the update Q table on standard output.
#
# Re-distribution for non-commercial purposes is permitted with attribution.


BEGIN {
	row = 0;
}

{
	if (row == 0) {
		# read in the parameters
		x_k   = $1;
		x_kp  = $2;
		u_k   = $3;
		r_k   = $4;
		alpha = $5;
		gamma = $6;

		printf("x_k=%s x_k+1=%s u_k=%s r_k=%f alpha=%f gamma=%f\n",
		       x_k, x_kp, u_k, r_k, alpha, gamma);

	} else {
		# read in a normal table row
		state[row]  = $1;
		action[row] = $2;
		Q[row]      = $3;
	}

	row ++;
}

END {

	# display initial table state
	printf("initial table:\n");
	printf("index\tstate\taction\tQ\n");
	maxQ = 0;
	maxu = "";
	for (i = 1 ; i < row ; i++) {

		printf("%i\t%s\t%s\t%i\n", i, state[i], action[i], Q[i]);

		# find max_u Q(x_k+1, u) as we go along
		if (state[i] == x_kp && Q[i] > maxQ) {

			maxQ = Q[i];
			maxu = action[i];
		}
	}
	printf("maxu=%s maxQ=%f\n", maxu, maxQ);

	# display updated table
	printf("\n\nupdated table:\n");
	printf("index\tstate\taction\tQ\n");
	for (i = 1 ; i < row ; i++) {

		if (state[i] == x_k && action[i] == u_k) {

			# compute the updated table row
			printf("%i\t%s\t%s\t", i, state[i], action[i]);

			# this is the implementation of the Q-update formula
			printf("%f\n", (1 + alpha) * Q[i] + alpha * (r_k - gamma * maxQ));

		} else {
			# row is unmodified
			printf("%i\t%s\t%s\t%i\n", i, state[i], action[i], Q[i]);
		}
	}
}
